// Group.cpp: Implementation of the CGroup class.
//
//////////////////////////////////////////////////////////////////////

#include "stdafx.h"
#include "Group.h"
#include "utils.h"
#include "pidl.h"
#include "lsapiex.h"
#include "main.h"
#include "DropTarget.h"

/**************************************************************************************************
	This is the constructor, it's called when the group is created.
**************************************************************************************************/
CGroup::CGroup(const char *szName) {
	IShellBrowserImpl();

	// Some defaults
	m_bInitialized = false;
	m_pDropTarget = NULL;
	m_nLastItem = -1;
	m_nBackFolderCount = 0;
	m_nForwardFolderCount = 0;
	wpOrigListView_Proc = NULL;

	utils::CopyString(&m_szName, szName); // Store the group name
	
	if (!ReadSettings(false)) // Read settings from step.rc
		return;

	m_hWnd = GetParentWindow(); // Get or create the parent window
	
	if (!InitFolderView()) // Set up the listview
		return;

	m_hwndNextClipViewer = SetClipboardViewer(m_hwndView); // Register for clipboard notifications

	m_bInitialized = true; // Let the main code know that the group loaded properly.
}

/**************************************************************************************************
	This is the destructor, it's called when the module is unloaded.
**************************************************************************************************/
CGroup::~CGroup() {
	SaveState(); // Save icon positions
	
	m_bInitialized = false; // Prevent onremove events from fireing.

	ChangeClipboardChain(m_hwndListView, m_hwndNextClipViewer); // Unregister from the windows clipboard chain

	if ( wpOrigListView_Proc != NULL )
        SetWindowLongPtr(m_hwndListView, GWLP_WNDPROC, (LONG) wpOrigListView_Proc); // Remove the subclass from the listview. 

	if (m_pDropTarget) {
		m_pDropTarget->Revoke(m_hwndView);
		delete m_pDropTarget;
	}
	if (m_pView2)
		m_pView2->Release();

	if (m_pView) {
		m_pView->DestroyViewWindow();
		m_pView->Release();
	}
	if (m_pFolder)
		m_pFolder->Release();

	if (m_hFont)
		DeleteObject(m_hFont);

	if (m_bDestroyParent)
		DestroyWindow(m_hWnd);

	if (m_szName)
		delete m_szName;
}

/**************************************************************************************************
	This function reads all the settings from step.rc and sets all the defaults.
**************************************************************************************************/
bool CGroup::ReadSettings(bool bIsRefresh) {
	char szBuf[MAX_LINE_LENGTH], szEvar[MAX_RCCOMMAND];

	SHGetSpecialFolderPath(NULL, m_szMyDocumentsPath, CSIDL_PERSONAL, false);
	SHGetSpecialFolderPath(NULL, m_szDesktopPath, CSIDL_DESKTOP, 0);

	if (!bIsRefresh) {
		LiteStep::GetPrefixedRCLine(szBuf, m_szName, "Folder", ".desktop");
		
		// Export CurrentFolder
		sprintf_s(szEvar, MAX_RCCOMMAND, "%s%s", m_szName, "CurrentFolder");

		if (_stricmp(szBuf, ".desktop") == 0) {
			strcpy_s(m_szFolderLocation, MAX_PATH, m_szDesktopPath);
			SHGetSpecialFolderLocation(NULL, CSIDL_DESKTOP, &m_pidlFolder);
			LSSetVariable(szEvar, ".desktop");
			//FixMyDocs();
		}
		else if (_stricmp(szBuf, ".mycomputer") == 0) {
			SHGetSpecialFolderPath(NULL, m_szFolderLocation, CSIDL_DRIVES, 0);
			SHGetSpecialFolderLocation(NULL, CSIDL_DRIVES, &m_pidlFolder);
			LSSetVariable(szEvar, ".mycomputer");
		}
		else {
			strcpy_s(m_szFolderLocation, szBuf);
			m_pidlFolder = PIDL::ParseDisplayName(szBuf);
			LSSetVariable(szEvar, m_szFolderLocation);
		}
		if (!m_pidlFolder)
			return false;
	}

	// This needs to be os dependant...
	strcpy_s(m_szRegDeskIconsPath, MAX_PATH, "Software\\Microsoft\\Windows\\CurrentVersion\\Explorer\\HideDesktopIcons\\NewStartPanel");

	LiteStep::GetPrefixedRCString(szBuf, m_szName, "ViewMode", "icons");
	if (_stricmp(szBuf, "tiles") == 0)
		m_uViewMode = FVM_TILE;
	else if (_stricmp(szBuf, "thumbnails") == 0)
		m_uViewMode = FVM_THUMBNAIL;
	else if (_stricmp(szBuf, "details") == 0)
		m_uViewMode = FVM_DETAILS;
	else if (_stricmp(szBuf, "smallicons") == 0)
		m_uViewMode = FVM_SMALLICON;
	else
		m_uViewMode = FVM_ICON;

	// Get default desktop icon settings.
	m_bHideRecycleBin = false;
	m_bHideMyDocuments = false;
	m_bHideInternetExplorer = false;
	m_bHideMyNetworkPlaces = false;
	m_bHideMyComputer = false;
	m_bDontUseRegistryIcons = LiteStep::GetPrefixedRCBool(m_szName, "DontUseRegistryIcons", FALSE);
	if (!m_bDontUseRegistryIcons)
		ReadIconSettings();
	m_bHideRecycleBin = LiteStep::GetPrefixedRCBool(m_szName, "HideRecycleBin", m_bHideRecycleBin);
	m_bHideMyDocuments = LiteStep::GetPrefixedRCBool(m_szName, "HideMyDocuments", m_bHideMyDocuments);
	m_bHideInternetExplorer = LiteStep::GetPrefixedRCBool(m_szName, "HideInternetExplorer", m_bHideInternetExplorer);
	m_bHideMyNetworkPlaces = LiteStep::GetPrefixedRCBool(m_szName, "HideMyNetworkPlaces", m_bHideMyNetworkPlaces);
	m_bHideMyComputer = LiteStep::GetPrefixedRCBool(m_szName, "HideMyComputer", m_bHideMyComputer);

	m_bUseIconPositionFile = true;
	LiteStep::GetPrefixedRCLine(m_szIconPositionFile, m_szName, "IconPositionFile", ".registry");
	if (_stricmp(m_szIconPositionFile, ".registry") == 0)
		m_bUseIconPositionFile = false;

	LiteStep::GetPrefixedRCString(szBuf, m_szName, "CustomDrive", "-");
	if (_stricmp(szBuf, ".current") == 0) {
		char lpCurrentDirectory[MAX_PATH];
		GetCurrentDirectory( MAX_PATH, lpCurrentDirectory );
		m_nCustomDrive = lpCurrentDirectory[0];
	}
	else 
		m_nCustomDrive = szBuf[0];

	LoadEvents();

	m_nX		= LiteStep::GetPrefixedRCCoordinate(m_szName, "X", 0, SCREEN_WIDTH);
	m_nY		= LiteStep::GetPrefixedRCCoordinate(m_szName, "Y", 0, SCREEN_HEIGHT);

	m_nWidth	= LiteStep::GetPrefixedRCInt(m_szName, "Width",  SCREEN_WIDTH);
	m_nHeight	= LiteStep::GetPrefixedRCInt(m_szName, "Height", SCREEN_HEIGHT);

	m_nSpacingX	= LiteStep::GetPrefixedRCInt(m_szName, "SpacingX",  -1);
	m_nSpacingY	= LiteStep::GetPrefixedRCInt(m_szName, "SpacingY", -1);
	
	m_bHidden	= LiteStep::GetPrefixedRCBool(m_szName, "StartHidden", FALSE);
	sprintf_s(szEvar, MAX_RCCOMMAND, "%s%s", m_szName, "IsHidden");
	LSSetVariable(szEvar, m_bHidden ? "true" : "false");

	m_bNoScroll	= LiteStep::GetPrefixedRCBool(m_szName, "NoScroll", FALSE);

	m_bDontPasteText = LiteStep::GetPrefixedRCBool(m_szName, "DontPasteText", FALSE);
	m_bDontPasteImages = LiteStep::GetPrefixedRCBool(m_szName, "DontPasteImages", FALSE);
	m_bDontPasteAudio = LiteStep::GetPrefixedRCBool(m_szName, "DontPasteAudio", FALSE);

	m_bDontUseRecycleBin	= LiteStep::GetPrefixedRCBool(m_szName, "DontUseRecycleBin", FALSE);
	m_bConfirmFileDelete	= LiteStep::GetPrefixedRCBool(m_szName, "ConfirmFileDelete", FALSE);
	m_bSimpleProgress		= LiteStep::GetPrefixedRCBool(m_szName, "SimpleProgress", FALSE);
	m_bNoConnectedFiles		= LiteStep::GetPrefixedRCBool(m_szName, "NoConnectedFiles", FALSE);
	m_bVirtualFileDelete	= LiteStep::GetPrefixedRCBool(m_szName, "VirtualFileDelete", FALSE);

	m_bExplorerDesktop = LiteStep::GetPrefixedRCBool(m_szName, "ExplorerDesktop", FALSE);

	m_wSelectionRectangleMod = GetModkeyWord("RectangleModifierKey", ".none");
	m_wSystemMenuMod = GetModkeyWord("SystemMenuModifierKey", ".none");

	m_bStealFocus	= LiteStep::GetPrefixedRCBool(m_szName, "StealFocus", FALSE);

	m_bNoInfoTips		= LiteStep::GetPrefixedRCBool(m_szName, "NoInfoTips", FALSE);
	m_bSnapToGrid		= LiteStep::GetPrefixedRCBool(m_szName, "SnapToGrid", TRUE);
	m_bHotTracking		= LiteStep::GetPrefixedRCBool(m_szName, "HotTracking", FALSE);
	m_nHotTrackingTime	= LiteStep::GetPrefixedRCInt(m_szName, "HotTrackingTime", -1);

	m_bInlineBrowsing = LiteStep::GetPrefixedRCBool(m_szName, "InlineBrowsing", FALSE);
	m_bNoVirtualSwitch = LiteStep::GetPrefixedRCBool(m_szName, "NoVirtualSwitch", FALSE);

	LiteStep::GetPrefixedRCString(szBuf, m_szName, "Background", "");
	utils::CopyString(&m_szBackground, szBuf);
	if (m_szBackground && m_szBackground[0]) {
		m_bTiled = LiteStep::GetPrefixedRCBool(m_szName, "Tiled", false);
		m_nOffsetPercentX = LiteStep::GetPrefixedRCInt(m_szName, "OffsetPercentX", 0);
		m_nOffsetPercentY = LiteStep::GetPrefixedRCInt(m_szName, "OffsetPercentY", 0);
	}
	m_crBackgroundColor= LiteStep::GetPrefixedRCColor(m_szName, "BackgroundColor", CLR_NONE);

	m_bAutoArrange = LiteStep::GetPrefixedRCBool(m_szName, "AutoArrange", FALSE);
	m_bAlignTop = LiteStep::GetPrefixedRCBool(m_szName, "AlignTop", FALSE);
	m_bAlwaysShowSelection =  LiteStep::GetPrefixedRCBool(m_szName, "AlwaysShowSelection", FALSE);

	// Get the font
	m_crTextColor= LiteStep::GetPrefixedRCColor(m_szName, "FontColor", RGB(255, 255, 255));
	m_crTextBackgroundColor= LiteStep::GetPrefixedRCColor(m_szName, "FontBackgroundColor", CLR_NONE);
	int nFontSize		= LiteStep::GetPrefixedRCInt(m_szName, "FontSize", 14);
	bool bFontBold		= LiteStep::GetPrefixedRCBool(m_szName, "FontBold", false);
	int nFontWeight		= LiteStep::GetPrefixedRCInt(m_szName, "FontWeight", bFontBold ? FW_BOLD : FW_NORMAL);
	bool bFontItalic	= LiteStep::GetPrefixedRCBool(m_szName, "FontItalic", false);
	bool bFontUnderline	= LiteStep::GetPrefixedRCBool(m_szName, "FontUnderline", false);
	bool bFontStrikeout	= LiteStep::GetPrefixedRCBool(m_szName, "FontStrikeout", false);
	int nFontQuality;
	LiteStep::GetPrefixedRCString(szBuf, m_szName, "FontQuality", "default");
	if (_stricmp(szBuf, "antialiased") == 0)
		nFontQuality = ANTIALIASED_QUALITY;
	else if (_stricmp(szBuf, "cleartype") == 0)
		nFontQuality = CLEARTYPE_QUALITY;
	else if (_stricmp(szBuf, "draft") == 0)
		nFontQuality = DRAFT_QUALITY;
	else if (_stricmp(szBuf, "nonantialiased") == 0)
		nFontQuality = NONANTIALIASED_QUALITY;
	else if (_stricmp(szBuf, "proof") == 0)
		nFontQuality = PROOF_QUALITY;
	else 
		nFontQuality = DEFAULT_QUALITY;
	LiteStep::GetPrefixedRCLine(szBuf, m_szName, "Font", "Arial");

	m_hFont = CreateFont(
		nFontSize,
		0, 0, 0,
		nFontWeight,
		bFontItalic,
		bFontUnderline,
		bFontStrikeout,
		DEFAULT_CHARSET,
		OUT_DEFAULT_PRECIS,
		CLIP_DEFAULT_PRECIS,
		nFontQuality,
		DEFAULT_PITCH | FF_DONTCARE,
		szBuf);
	if (!m_hFont)
		utils::ErrorBoxPrint("Could not load font %s", szBuf);

	return true;
}

/**************************************************************************************************
	This function should hide the items we don't want to display on the desktop. It's called
	when an item is inserted on the desktop. iItem is the id of the inserted item. This function
	returns true if the item was added and false if it was removed.
**************************************************************************************************/
bool CGroup::HandleInsItem(int iItem) {
	char szFileName[MAX_PATH_LENGTH];

	GetNameFromId(iItem, szFileName);

	if (!IconShouldBeHidden(szFileName))
		return true;

	ListView_DeleteItem(m_hwndListView, iItem);
	return false;
}

/**************************************************************************************************
	This function scans all desktop icons and hides the ones that should be hidden.
**************************************************************************************************/
void CGroup::ScanIcons() {
	char szFileName[MAX_PATH_LENGTH];
	int nCount = ListView_GetItemCount(m_hwndListView);

	for (int i = 0; i < nCount; i++) {
		GetNameFromId(i, szFileName);
		if (IconShouldBeHidden(szFileName)) {
			ListView_DeleteItem(m_hwndListView, i); // Delete the item
			nCount--; i--; // the list reorders itself after every deletion
			if (m_bInitialized && !m_bInFolderChange) // Execute OnRemoveItem if the folder is ready
				HandleEvent(EVENT_ONREMOVEITEM);
		}
	}
}

/**************************************************************************************************
	This function determines if an icon should be hidden or not.
**************************************************************************************************/
bool CGroup::IconShouldBeHidden(const char *szFileName) {
	if (_stricmp(m_szFolderLocation, m_szDesktopPath) != 0)
		return false;
	if (m_bHideMyDocuments && !strcmp(szFileName, m_szMyDocumentsPath))
		return true;
	else if (m_bHideInternetExplorer && !strcmp(szFileName, "::{871C5380-42A0-1069-A2EA-08002B30309D}"))
		return true;
	else if (m_bHideMyNetworkPlaces && !strcmp(szFileName, "::{208D2C60-3AEA-1069-A2D7-08002B30309D}"))
		return true;
	else if (m_bHideMyComputer && !strcmp(szFileName, "::{20D04FE0-3AEA-1069-A2D8-08002B30309D}"))
		return true;
	else if (m_bHideRecycleBin && !strcmp(szFileName, "::{645FF040-5081-101B-9F08-00AA002F954E}"))
		return true;
	return false;
}

/**************************************************************************************************
	This function shows/hides virtual desktop icons.
**************************************************************************************************/
void CGroup::ToggleIcon(const char *szFileName, bool bVisible, bool bAlreadyDeleted) {
	DWORD dwRegSetting = bVisible ? 0 : 1;
	if (!strcmp(szFileName, m_szMyDocumentsPath)) {
		m_bHideMyDocuments = !bVisible;
		if (!m_bDontUseRegistryIcons) {
			SHSetValue(HKEY_CURRENT_USER, m_szRegDeskIconsPath, "{450D8FBA-AD25-11D0-98A8-0800361B1103}", REG_DWORD, &dwRegSetting, sizeof(dwRegSetting));
			SendMessage (HWND_BROADCAST, WM_SETTINGCHANGE, 0, (LPARAM)"Windows");
		}
	}
	else if (!strcmp(szFileName, "::{871C5380-42A0-1069-A2EA-08002B30309D}")) {
		m_bHideInternetExplorer = !bVisible;
		if (!m_bDontUseRegistryIcons) {
			SHSetValue(HKEY_CURRENT_USER, m_szRegDeskIconsPath, "{871C5380-42A0-1069-A2EA-08002B30309D}", REG_DWORD, &dwRegSetting, sizeof(dwRegSetting));
			SendMessage (HWND_BROADCAST, WM_SETTINGCHANGE, 0, (LPARAM)"Windows");
		}
	}
	else if (!strcmp(szFileName, "::{208D2C60-3AEA-1069-A2D7-08002B30309D}")) {
		m_bHideMyNetworkPlaces = !bVisible;
		if (!m_bDontUseRegistryIcons) {
			SHSetValue(HKEY_CURRENT_USER, m_szRegDeskIconsPath, "{208D2C60-3AEA-1069-A2D7-08002B30309D}", REG_DWORD, &dwRegSetting, sizeof(dwRegSetting));
			SendMessage (HWND_BROADCAST, WM_SETTINGCHANGE, 0, (LPARAM)"Windows");
		}
	}
	else if (!strcmp(szFileName, "::{20D04FE0-3AEA-1069-A2D8-08002B30309D}")) {
		m_bHideMyComputer = !bVisible;
		if (!m_bDontUseRegistryIcons) {
			SHSetValue(HKEY_CURRENT_USER, m_szRegDeskIconsPath, "{20D04FE0-3AEA-1069-A2D8-08002B30309D}", REG_DWORD, &dwRegSetting, sizeof(dwRegSetting));
			SendMessage (HWND_BROADCAST, WM_SETTINGCHANGE, 0, (LPARAM)"Windows");
		}
	}
	else if (!strcmp(szFileName, "::{645FF040-5081-101B-9F08-00AA002F954E}")) {
		m_bHideRecycleBin = !bVisible;
		if (!m_bDontUseRegistryIcons) {
			SHSetValue(HKEY_CURRENT_USER, m_szRegDeskIconsPath, "{645FF040-5081-101B-9F08-00AA002F954E}", REG_DWORD, &dwRegSetting, sizeof(dwRegSetting));
			SendMessage (HWND_BROADCAST, WM_SETTINGCHANGE, 0, (LPARAM)"Windows");
		}
	}
	else
		return; // Don't bother with the rest of the code if none of the targeted filenames were found.

	if (_stricmp(m_szDesktopPath, m_szFolderLocation) != 0)
		return; // The rest of the code is just to refresh the desktop folder.

	if (bVisible) {
		/*	We need to rescan the directory somehow in order to show previously hidden icons.
			Maybe we could store all LVITEM structures and use ListView_InsertItem, but this works fine
			for now. */
		SaveState();
		m_pView->Refresh();
		RestoreState();
	}
	else {
		if (!bAlreadyDeleted)
			ScanIcons(); // Remove hidden icons.
	}
}

/**************************************************************************************************
	This function reads windows desktop icon preferences from the registry.
**************************************************************************************************/
void CGroup::ReadIconSettings() {
	DWORD dwBuffer, dwSize = sizeof(DWORD), dwType = REG_DWORD;
	SHGetValue(HKEY_CURRENT_USER, m_szRegDeskIconsPath, "{871C5380-42A0-1069-A2EA-08002B30309D}", &dwType, &dwBuffer, &dwSize);
	m_bHideInternetExplorer = (dwBuffer == 1);
	SHGetValue(HKEY_CURRENT_USER, m_szRegDeskIconsPath, "{20D04FE0-3AEA-1069-A2D8-08002B30309D}", &dwType, &dwBuffer, &dwSize);
	m_bHideMyComputer = (dwBuffer == 1);
	SHGetValue(HKEY_CURRENT_USER, m_szRegDeskIconsPath, "{450D8FBA-AD25-11D0-98A8-0800361B1103}", &dwType, &dwBuffer, &dwSize);
	m_bHideMyDocuments = (dwBuffer == 1);
	SHGetValue(HKEY_CURRENT_USER, m_szRegDeskIconsPath, "{208D2C60-3AEA-1069-A2D7-08002B30309D}", &dwType, &dwBuffer, &dwSize);
	m_bHideMyNetworkPlaces = (dwBuffer == 1);
	SHGetValue(HKEY_CURRENT_USER, m_szRegDeskIconsPath, "{645FF040-5081-101B-9F08-00AA002F954E}", &dwType, &dwBuffer, &dwSize);
	m_bHideRecycleBin = (dwBuffer == 1);
}

/**************************************************************************************************
	Sorting routines...
**************************************************************************************************/
int SortPIDLs(LPSHELLFOLDER psf, LPARAM lColumn, LPITEMIDLIST pidl1, LPITEMIDLIST pidl2) {
	HRESULT hres = psf->CompareIDs(lColumn, pidl1, pidl2);
	return static_cast<signed short>(HRESULT_CODE(hres));
}
int CALLBACK SortByNameCompareFunc(LPARAM lParam1, LPARAM lParam2, LPARAM lParamSort) {
	return SortPIDLs(
		reinterpret_cast<LPSHELLFOLDER>(lParamSort), 
		0,
		reinterpret_cast<LPITEMIDLIST>(lParam1),
		reinterpret_cast<LPITEMIDLIST>(lParam2)
	);
}
int CALLBACK SortBySizeCompareFunc(LPARAM lParam1, LPARAM lParam2, LPARAM lParamSort) {
	return SortPIDLs(
		reinterpret_cast<LPSHELLFOLDER>(lParamSort), 
		1,
		reinterpret_cast<LPITEMIDLIST>(lParam1),
		reinterpret_cast<LPITEMIDLIST>(lParam2)
	);
}
int CALLBACK SortByTypeCompareFunc(LPARAM lParam1, LPARAM lParam2, LPARAM lParamSort) {
	return SortPIDLs(
		reinterpret_cast<LPSHELLFOLDER>(lParamSort), 
		2,
		reinterpret_cast<LPITEMIDLIST>(lParam1),
		reinterpret_cast<LPITEMIDLIST>(lParam2)
	);
}
int CALLBACK SortByDateCompareFunc(LPARAM lParam1, LPARAM lParam2, LPARAM lParamSort) {
	return SortPIDLs(
		reinterpret_cast<LPSHELLFOLDER>(lParamSort), 
		3,
		reinterpret_cast<LPITEMIDLIST>(lParam1),
		reinterpret_cast<LPITEMIDLIST>(lParam2)
	);
}

/**************************************************************************************************
	This function moves and resizes the group
**************************************************************************************************/
void CGroup::Reposition (int x, int y, int width, int height, bool relative) {
	if (relative) {
		m_nX += x;
		m_nY += y;
		m_nWidth += width;
		m_nHeight += height;
	}
	else {
		m_nX = x;
		m_nY = y;
		m_nWidth = width;
		m_nHeight = height;
	}
	if (m_bDestroyParent)
		MoveWindow(m_hWnd, m_nX, m_nY, m_nWidth, m_nHeight, true);
	SaveState();
	if (m_pView2)
		m_pView2->Release();
	if (m_pView) {
		m_pView->DestroyViewWindow();
		m_pView->Release();
	}
	InitFolderView();
}

/**************************************************************************************************
	Hook function?
**************************************************************************************************/
/*void CGroup::Hook(HWND hParent) {
	if (hParent != GetParent(m_hwndView)) {
		SetWindowLongPtr(m_hwndView, GWL_STYLE, (GetWindowLongPtr(m_hwndView, GWL_STYLE) & ~WS_POPUP)|WS_CHILD);
		SetParent(m_hwndView, hParent);
	}
}*/

/**************************************************************************************************
	This function checks if a set of modifier keys are pressed.
**************************************************************************************************/
bool CGroup::CheckModkeys(WORD wKeys) {
	if (((wKeys & MODKEY_SHIFT) == MODKEY_SHIFT) && !(GetKeyState( VK_SHIFT ) < 0))
		return false;
	if (((wKeys & MODKEY_ALT) == MODKEY_ALT) && !(GetKeyState( VK_MENU ) < 0))
		return false;
	if (((wKeys & MODKEY_CTRL) == MODKEY_CTRL) && !(GetKeyState( VK_CONTROL ) < 0))
		return false;
	return true;
}

/**************************************************************************************************
	This function changes the directory
**************************************************************************************************/
void CGroup::ChangeDir(const char *szNewDir) {
	char szCDPath[MAX_LINE_LENGTH], szEvar[MAX_RCCOMMAND];
	bool bSetFocus = (GetFocus() == m_hwndListView);

	// Determine what folder to browse to and if the folder is real.
	strcpy_s(szCDPath, szNewDir);

	if (_stricmp(szCDPath, ".back") == 0) {
		if (m_nBackFolderCount == 0)
			return;

		m_nBackFolderCount--;
		m_nForwardFolderCount++;
	}
	if (_stricmp(szCDPath, ".forward") == 0) {
		if (m_nForwardFolderCount == 0)
			return;

		m_nForwardFolderCount--;
		m_nBackFolderCount++;
	}

	if (_stricmp(szCDPath, ".selected") == 0) {
		int iItem = ListView_GetNextItem(m_hwndListView, -1, LVNI_FOCUSED);
		if (iItem == -1)
			return;
		GetNameFromId(iItem, szCDPath);
		if (!utils::Is_Directory(szCDPath))
			return;
	}
	if (_stricmp(szCDPath, ".up") == 0) {
		int iLast = -1, iCnt = 0, s = sizeof(m_szFolderLocation), i;
		for (i = 0; i <= s; i++) {
			if (m_szFolderLocation[i] == 92) {
				iLast = i;
				iCnt++;
			}
		}
		if (iCnt == 1) { // This takes some special managment...
			if (m_szFolderLocation[3] == 0)
				return; // We are already at the root drive.
			else
				iLast++; // We want to keep the \ at the end of the path.
		}
		memcpy(szCDPath, m_szFolderLocation, sizeof(m_szFolderLocation));
		szCDPath[iLast] = 0; // End the string there.
	}
	if (!m_bNoVirtualSwitch && !_stricmp(szCDPath, m_szDesktopPath))
		strcpy_s(szCDPath, MAX_LINE_LENGTH, ".desktop");
	else if (!utils::Is_Directory(szCDPath)) {
		if ((_stricmp(szCDPath, ".desktop") != 0) && (_stricmp(szCDPath, ".mycomputer") != 0)) {
			utils::ErrorBoxPrint("Invalid Directory: %s", szNewDir);
			return;
		}
	}

	// Destroy the current folder view.
	m_bInFolderChange = true; // No input while changing folders.
	SaveState();
	if (m_pDropTarget) {
		m_pDropTarget->Revoke(m_hwndView);
		delete m_pDropTarget;
	}
	if (m_pView2)
		m_pView2->Release();
	if (m_pView) {
		m_pView->DestroyViewWindow();
		m_pView->Release();
	}
	if (m_pFolder)
		m_pFolder->Release();

	// Load the contents of the new folder.
	if (_stricmp(szCDPath, ".desktop") == 0) {
		strcpy_s(m_szFolderLocation, MAX_PATH, m_szDesktopPath);
		SHGetSpecialFolderLocation(NULL, CSIDL_DESKTOP, &m_pidlFolder);
		//FixMyDocs();
	}
	else if (_stricmp(szCDPath, ".mycomputer") == 0) {
		SHGetSpecialFolderPath(NULL, m_szFolderLocation, CSIDL_DRIVES, 0);
		SHGetSpecialFolderLocation(NULL, CSIDL_DRIVES, &m_pidlFolder);
	}
	else {
		strcpy_s(m_szFolderLocation, szCDPath);
		m_pidlFolder = PIDL::ParseDisplayName(szCDPath);
	}

	// Export CurrentFolder
	sprintf_s(szEvar, MAX_RCCOMMAND, "%s%s", m_szName, "CurrentFolder");
	LSSetVariable(szEvar, szCDPath);

	InitFolderView();// Recreate the folder view.
	if (bSetFocus)
		SetFocus(m_hwndListView);

	HandleEvent(EVENT_ONFOLDERCHANGE); // Execute OnFolderChange
	m_bInFolderChange = false; // Done changing folders, unlock the listview procedure
}

/**************************************************************************************************
	This function changes the video mode
**************************************************************************************************/
void CGroup::SetViewMode(const char *szNewMode) {
	UINT uNewMode;

	if (_stricmp(szNewMode, "tiles") == 0)
		uNewMode = FVM_TILE;
	else if (_stricmp(szNewMode, "thumbnails") == 0)
		uNewMode = FVM_THUMBNAIL;
	else if (_stricmp(szNewMode, "details") == 0)
		uNewMode = FVM_DETAILS;
	else if (_stricmp(szNewMode, "icons") == 0)
		uNewMode = FVM_ICON;
	else if (_stricmp(szNewMode, "smallicons") == 0)
		uNewMode = FVM_SMALLICON;
	else
		return;

	if (uNewMode == m_uViewMode)
		return;

	m_uViewMode = uNewMode;

	// ListView_SetView doesn't do that good of a job
	SaveState();
	if (m_pView2)
		m_pView2->Release();
	if (m_pView) {
		m_pView->DestroyViewWindow();
		m_pView->Release();
	}
	InitFolderView();
}

/**************************************************************************************************
	This function changes the hottracking setting
**************************************************************************************************/
void CGroup::HotTracking( bool bEnabled ) {
	if (bEnabled) {
		ListView_SetExtendedListViewStyleEx(m_hwndListView, LVS_EX_TRACKSELECT, LVS_EX_TRACKSELECT);
		m_bHotTracking = true;
	}
	else {
		ListView_SetExtendedListViewStyleEx(m_hwndListView, LVS_EX_TRACKSELECT, 0);
		m_bHotTracking = false;
	}
	
}

/**************************************************************************************************
	This function changes the infotip setting
**************************************************************************************************/
void CGroup::InfoTip( bool bEnabled ) {
	if (bEnabled) {
		ListView_SetExtendedListViewStyleEx(m_hwndListView, LVS_EX_INFOTIP, LVS_EX_INFOTIP);
		m_bNoInfoTips = false;
	}
	else {
		ListView_SetExtendedListViewStyleEx(m_hwndListView, LVS_EX_INFOTIP, 0);
		m_bNoInfoTips = true;
	}	
}

/**************************************************************************************************
	This function changes the aligntop setting
**************************************************************************************************/
void CGroup::AlignTop( bool bEnabled ) {
	if (bEnabled) {
		if (!m_bAlignTop) {
			m_longListViewStyle ^= LVS_ALIGNLEFT;
			m_bAlignTop = true;
		}
	}
	else {
		if (m_bAlignTop) {
			m_longListViewStyle |= LVS_ALIGNLEFT;
			m_bAlignTop = false;
		}
	}
	SetWindowLongPtr(m_hwndListView, GWL_STYLE, m_longListViewStyle);
}


/**************************************************************************************************
	This function changes the snap to grid setting
**************************************************************************************************/
void CGroup::SnapToGrid( bool bEnabled ) {
	if (bEnabled) {
		m_bSnapToGrid = true;
		ListView_SetExtendedListViewStyleEx(m_hwndListView, LVS_EX_SNAPTOGRID, LVS_EX_SNAPTOGRID);
		ListView_Arrange(m_hwndListView, LVA_SNAPTOGRID);
	}
	else {
		m_bSnapToGrid = false;
		ListView_SetExtendedListViewStyleEx(m_hwndListView, LVS_EX_SNAPTOGRID, 0);
	}
}

/**************************************************************************************************
	This function resets stored icon positions
**************************************************************************************************/
void CGroup::Reset(const char *szFolder, bool bDontSort) {
	char szKeyName[MAX_PATH], szNum[10], szFile[MAX_PATH];
	HKEY hKey;
	
	if (_stricmp(szFolder, ".all") == 0) { 
		if (m_bUseIconPositionFile)
			remove(m_szIconPositionFile); // Just delete the file
		else {
			StringCchPrintf(szKeyName, MAX_PATH, "%s%s", REGKEY_MODULE, m_szName);
			SHDeleteKey(HKEY_CURRENT_USER, szKeyName);
		}
		if (!bDontSort)
			ListView_SortItems(m_hwndListView, SortByNameCompareFunc, m_pFolder);
	}
	else if (_stricmp(szFolder, "") == 0) {
		if (m_bUseIconPositionFile) {
			strcpy_s(m_szFolderLocation, szFile);
			if (szFile[0] == m_nCustomDrive)
				szFile[0] = 63;
			WritePrivateProfileSection(szFile, "", m_szIconPositionFile);
		}
		else {
			StringCchPrintf(szKeyName, MAX_PATH, "%s%s\\%s", REGKEY_MODULE, m_szName, m_szFolderLocation);
			if (ERROR_SUCCESS != RegOpenKeyEx( HKEY_CURRENT_USER, szKeyName, 0, KEY_ALL_ACCESS, &hKey))
				return;
			for (int k = 0; ; k++) {
				StringCchPrintf(szNum, 10, "%d", k);
				if (RegDeleteValue(hKey, szNum) != ERROR_SUCCESS)
					break;
			}
			RegCloseKey(hKey);
		}
		if (!bDontSort)
			ListView_SortItems(m_hwndListView, SortByNameCompareFunc, m_pFolder);
	}
	else {
		if (m_bUseIconPositionFile) {
			memcpy(szFile, szFolder, sizeof(szFolder));
			if (szFile[0] == m_nCustomDrive)
				szFile[0] = 63;
			WritePrivateProfileSection(szFile, "", m_szIconPositionFile);
		}
		else {
			StringCchPrintf(szKeyName, MAX_PATH, "%s%s\\%s", REGKEY_MODULE, m_szName, szFolder);
			if (ERROR_SUCCESS != RegOpenKeyEx( HKEY_CURRENT_USER, szKeyName, 0, KEY_ALL_ACCESS, &hKey))
				return;
			for (int k = 0; ; k++) {
				StringCchPrintf(szNum, 10, "%d", k);
				if (RegDeleteValue(hKey, szNum) != ERROR_SUCCESS)
					break;
			}
			RegCloseKey(hKey);
		}
		if (_stricmp(szFolder, m_szFolderLocation) == 0 && !bDontSort)
			ListView_SortItems(m_hwndListView, SortByNameCompareFunc, m_pFolder);
	}
}

/**************************************************************************************************
	This function deletes the selected files
**************************************************************************************************/
void CGroup::DeleteFiles() {
	char szFileName[MAX_PATH], szFullCommand[MAX_LINE_LENGTH] = {0};
	int iItem = -1;
	UINT i, j;
	UINT iNumSelected = ListView_GetSelectedCount(m_hwndListView);
	UINT iCommandPos = 0;
	UINT iFakeFiles[256], iRealFiles[256];
	UINT iFakeFileCounter = 0, iRealFileCounter = 0;
	SHFILEOPSTRUCT SHFileDelete;
	STRRET strDispName;
	LVITEM lvitem;
	size_t FileNameSize;

	if (iNumSelected == 0)
		return;

	if (m_bVirtualFileDelete)
		return DeleteFilesVirtually();

	for (i = 0; i < iNumSelected; i++) {
		iItem = ListView_GetNextItem(m_hwndListView, iItem, LVNI_SELECTED);
		ZeroMemory(&lvitem, sizeof(LVITEM));
		lvitem.mask = LVIF_PARAM;
		lvitem.iItem = iItem;
		lvitem.iSubItem = 0;
		ListView_GetItem(m_hwndListView, &lvitem);
		m_pFolder->GetDisplayNameOf((LPCITEMIDLIST)(lvitem.lParam), SHGDN_FORPARSING, &strDispName);
		StrRetToBuf(&strDispName, (LPCITEMIDLIST)(lvitem.lParam), szFileName, MAX_PATH);
		StringCchLength(szFileName, MAX_PATH, &FileNameSize);
		if (utils::File_Exists(szFileName)) {
			iRealFiles[iRealFileCounter] = iItem;
			iRealFileCounter++;
			for(j = 0; j <= FileNameSize; j++) {
				szFullCommand[iCommandPos++] = szFileName[j];
			}
		}
		else {
			iFakeFiles[iFakeFileCounter] = iItem;
			iFakeFileCounter++;
		}
	}
	for (i = 0; i < iFakeFileCounter; i++) {
		for (j = 0; j < iRealFileCounter; j++) {
			if (iRealFiles[j] < iFakeFiles[i])
				iFakeFiles[i]--;
		}
	}
	if (iRealFileCounter == 0) // There are no real files selected
		return DeleteFilesVirtually();

	ZeroMemory(&SHFileDelete, sizeof(SHFILEOPSTRUCT));
	SHFileDelete.wFunc = FO_DELETE;
	SHFileDelete.hwnd = NULL;
	SHFileDelete.pFrom = szFullCommand;
	SHFileDelete.pTo = NULL;
	SHFileDelete.fFlags = 0;
	if (m_bSimpleProgress)
		SHFileDelete.fFlags |= FOF_SIMPLEPROGRESS;
	if (!m_bDontUseRecycleBin)
		SHFileDelete.fFlags |= FOF_ALLOWUNDO;
	if (!m_bConfirmFileDelete)
		SHFileDelete.fFlags |= FOF_NOCONFIRMATION | FOF_WANTNUKEWARNING;
	if (m_bNoConnectedFiles)
		SHFileDelete.fFlags |= FOF_NO_CONNECTED_ELEMENTS;
	if (SHFileOperation(&SHFileDelete) == 0) {
		if (!SHFileDelete.fAnyOperationsAborted) {
			for (i = 0; i < iFakeFileCounter; i++) {
				GetNameFromId(iFakeFiles[i], szFileName);
				ToggleIcon(szFileName, false, true);
			}
		}
	}
}

/**************************************************************************************************
	This function removes files from the view
**************************************************************************************************/
void CGroup::DeleteFilesVirtually() {
	// We need to make this function keep the removed icons removed.
	char szFileName[256], szTemp[512];
	int nSelected = ListView_GetSelectedCount(m_hwndListView), iItem = -1, i;
	if (nSelected == 1) {
		ListView_GetItemText(m_hwndListView, ListView_GetNextItem(m_hwndListView, iItem, LVNI_SELECTED), 0, szFileName, 256);
		sprintf_s(szTemp, "Are you sure that you want to delete %s?", szFileName);
	}
	else
		sprintf_s(szTemp, "Are you sure that you want to delete these %i files?", nSelected);

	if (MessageBox(NULL, szTemp, "Confirm File Delete", MB_YESNO | MB_ICONQUESTION) == IDYES) {	
		for (i = 0; i < nSelected; i++) {
			iItem = ListView_GetNextItem(m_hwndListView, iItem, LVNI_SELECTED);
			GetNameFromId(iItem, szFileName);
			ListView_DeleteItem(m_hwndListView, iItem);
			ToggleIcon(szFileName, false, true);
		}
	}
}
/**************************************************************************************************
	Debug function
**************************************************************************************************/
/*void CGroup::EchoFile() {
	char szFileName[MAX_PATH];
	int iItem = -1;
	STRRET strDispName;
	LVITEM lvitem;
	size_t FileNameSize = 0;
	SFGAOF flags;

	iItem = ListView_GetNextItem(m_hwndListView, iItem, LVNI_SELECTED);
	ZeroMemory(&lvitem, sizeof(LVITEM));
	lvitem.mask = LVIF_PARAM;
	lvitem.iItem = iItem;
	lvitem.iSubItem = 0;
	ListView_GetItem(m_hwndListView, &lvitem);
	m_pFolder->GetDisplayNameOf((LPCITEMIDLIST)(lvitem.lParam), SHGDN_FORPARSING, &strDispName);
	flags = SFGAO_FILESYSTEM;
	m_pFolder->GetAttributesOf(1, (LPCITEMIDLIST*)(&lvitem.lParam), &flags);
	StrRetToBuf(&strDispName, (LPCITEMIDLIST)(lvitem.lParam), szFileName, MAX_PATH);
	StringCchLength(szFileName, MAX_PATH, &FileNameSize);

	sprintf_s(szFileName, MAX_PATH, "%s", lvitem.lParam);

	//StringFromCLSID(lvitem.lParam, &szFileName);

	for (int i = 0; i < sizeof(szFileName); i++)
		if (szFileName[i] == 0)
			szFileName[i] = 63;

	//if ((flags & SFGAO_FILESYSTEM) != SFGAO_FILESYSTEM)
	//	MessageBox(NULL, "It's not a filesystem object", "Debug - Info", MB_OK);

	MessageBox(NULL, szFileName, "Debug - Filename", MB_OK);
}*/

/**************************************************************************************************
	Clears the cutmark when the clipboard changes
**************************************************************************************************/
void CGroup::ClearCutMark() {
	if (!m_bCutMarkedFiles)
		return;
	ListView_SetItemState(m_hwndListView, -1, 0, LVIS_CUT);
	m_bCutMarkedFiles = false;
}

/**************************************************************************************************
	This function puts selected icons in the clipboard
**************************************************************************************************/
void CGroup::CopyFiles(bool cut) {
	char szFileName[MAX_PATH_LENGTH], szFullCommand[MAX_LINE_LENGTH*2];
	int iItem = -1;
	UINT i, j, iCommandPos = 0, iFiles[256], iFileCounter = 0;
	UINT iNumSelected = ListView_GetSelectedCount(m_hwndListView);
	size_t FileNameSize = 0;
	
	if (!(iNumSelected > 0))
		return; // No files selected

	ClearCutMark();

	// Build the list of files
	for (i = 0; i < iNumSelected; i++) {
		iItem = ListView_GetNextItem(m_hwndListView, iItem, LVNI_SELECTED);
		GetNameFromId(iItem, szFileName);
		StringCchLength(szFileName, MAX_PATH, &FileNameSize);
		if (utils::File_Exists(szFileName)) { // We ignore virtual files.
			if (cut) 
				ListView_SetItemState(m_hwndListView, iItem, LVIS_CUT, LVIS_CUT);
			iFiles[iFileCounter] = iItem;
			iFileCounter++;
			for(j = 0; j < FileNameSize; j++) {
				szFullCommand[iCommandPos++] = szFileName[j];
			}
			szFullCommand[iCommandPos++] = 0;
		}
	}
	szFullCommand[iCommandPos++] = 0;

	if (!(iFileCounter > 0))
		return; // In case they tried to copy my computer or something like that

	DROPFILES dobj = { 20, { 0, 0 }, 0, 1 };
	HGLOBAL hClipData = GlobalAlloc( GMEM_ZEROINIT | GMEM_MOVEABLE | GMEM_DDESHARE, sizeof(dobj) + iCommandPos*2 + 5 );
	char* szData = (char*)GlobalLock(hClipData);
	memcpy( szData, &dobj, 20 );
	char* sWStr = szData+20;
	for ( i = 0; i < iCommandPos*2; i += 2 )
		sWStr[i] = szFullCommand[i/2];
	GlobalUnlock(hClipData);
	if (OpenClipboard(NULL)) {
		EmptyClipboard();
		SetClipboardData(CF_HDROP, hClipData);
		HGLOBAL h2 = GlobalAlloc(GMEM_ZEROINIT|GMEM_MOVEABLE|GMEM_DDESHARE, sizeof(DWORD) );
		DWORD* pdw1 = (DWORD*)GlobalLock(h2);
		*pdw1 = cut ? DROPEFFECT_MOVE : DROPEFFECT_COPY;
		GlobalUnlock(h2);
		SetClipboardData( RegisterClipboardFormat( CFSTR_PREFERREDDROPEFFECT ), h2 );
		CloseClipboard();
	}
	m_bCutMarkedFiles = true;
}

/**************************************************************************************************
	This function determines the type of data in the clipboard and calls the appropriate function
**************************************************************************************************/
void CGroup::DoPaste() {
	if (IsClipboardFormatAvailable(CF_HDROP))
		return PasteFiles();
	if (IsClipboardFormatAvailable(CF_TEXT))
		return PasteText();
	if (IsClipboardFormatAvailable(CF_BITMAP))
		return PasteBitmap();
	if (IsClipboardFormatAvailable(CF_WAVE))
		return PasteWave();
}
/**************************************************************************************************
	This function pastes files in the clipboard into the group's folder
**************************************************************************************************/
void CGroup::PasteFiles() {
	char *files, szFiles[MAX_LINE_LENGTH*2] = {0};
	bool bMove = false, bBreak = false;
	DWORD dropeffect;
	UINT cfDropEffect, i = 0;
	SHFILEOPSTRUCT SHFilePaste;
	
	if (!OpenClipboard(NULL))
		return;

	// Build the pFrom string.
	files = (char*)GetClipboardData(CF_HDROP)+20;
	while (!bBreak) {
		szFiles[i/2] = files[i];
		if (files[i] == 0 && files[i+2]== 0)
			bBreak = true;
		i += 2;
	}

	// Check if the file should be moved rather than copied.
	cfDropEffect = RegisterClipboardFormat(CFSTR_PREFERREDDROPEFFECT);
	if (IsClipboardFormatAvailable(cfDropEffect)) {
		// This won't work it litestep is hosting the folder window
		memcpy(&dropeffect, GetClipboardData(cfDropEffect), sizeof(DWORD));
		if ((dropeffect & DROPEFFECT_MOVE) == DROPEFFECT_MOVE)
			bMove = true;
	}

	ZeroMemory(&SHFilePaste, sizeof(SHFILEOPSTRUCT));
	SHFilePaste.wFunc = bMove ? FO_MOVE : FO_COPY;
	SHFilePaste.hwnd = NULL;
	SHFilePaste.pFrom = szFiles;
	SHFilePaste.pTo = m_szFolderLocation;
	SHFilePaste.fFlags = FOF_NOCONFIRMMKDIR;
	if (m_bSimpleProgress)
		SHFilePaste.fFlags |= FOF_SIMPLEPROGRESS;
	if (m_bNoConnectedFiles)
		SHFilePaste.fFlags |= FOF_NO_CONNECTED_ELEMENTS;
	SHFileOperation(&SHFilePaste);
	if (bMove && !SHFilePaste.fAnyOperationsAborted)
		EmptyClipboard(); // Perhaps we should msg explorer to do this for us, but why?
	CloseClipboard();
}

/**************************************************************************************************
	This function pastes text in the clipboard into the group's folder
**************************************************************************************************/
void CGroup::PasteText() {
	if (m_bDontPasteText)
		return;
	if (!OpenClipboard(NULL))
		return;

	FILE* pFile;
	char* szText = (char*)GetClipboardData(CF_TEXT);
	char szFileLoc[MAX_PATH_LENGTH];

	SYSTEMTIME st;
	GetSystemTime(&st);
	sprintf_s(szFileLoc, MAX_PATH_LENGTH, "%s\\clipboard-%d-%d-%d-%d-%d-%d-%d.txt", m_szFolderLocation,
		st.wYear, st.wMonth, st.wDay, st.wHour, st.wMinute, st.wSecond, st.wMilliseconds);
	fopen_s (&pFile, szFileLoc, "a");
	fputs (szText, pFile);
	fclose (pFile);

	CloseClipboard();
}

/**************************************************************************************************
	This function pastes bitmaps in the clipboard into the group's folder
**************************************************************************************************/
void CGroup::PasteBitmap() {
	if (m_bDontPasteImages)
		return;
	/*if (!OpenClipboard(NULL))
		return;

	FILE* pFile;
	HBITMAP hBMP = (HBITMAP)GetClipboardData(CF_BITMAP);
	char szFileLoc[MAX_PATH_LENGTH];

	SYSTEMTIME st;
	GetSystemTime(&st);
	sprintf_s(szFileLoc, MAX_PATH_LENGTH, "%s\\clipboard-%d-%d-%d-%d-%d-%d-%d.bmp", m_szFolderLocation,
		st.wYear, st.wMonth, st.wDay, st.wHour, st.wMinute, st.wSecond, st.wMilliseconds);
	fopen_s (&pFile, szFileLoc, "a");
	fputs (hBMP, pFile);
	fclose (pFile);

	CloseClipboard();*/
}

/**************************************************************************************************
	This function pastes waves in the clipboard into the group's folder
**************************************************************************************************/
void CGroup::PasteWave() {
	if (m_bDontPasteAudio)
		return;
}

/**************************************************************************************************
	Sorting functions called by bangs. They need to be in CGroup
**************************************************************************************************/
void CGroup::SortByName() {
	ListView_SortItems(m_hwndListView, SortByNameCompareFunc, m_pFolder);
}

void CGroup::SortBySize() {
	ListView_SortItems(m_hwndListView, SortBySizeCompareFunc, m_pFolder);
}

void CGroup::SortByType() {
	ListView_SortItems(m_hwndListView, SortByTypeCompareFunc, m_pFolder);
}

void CGroup::SortByDate() {
	ListView_SortItems(m_hwndListView, SortByDateCompareFunc, m_pFolder);
}

void CGroup::SetTextColor(COLORREF crColor) {
	ListView_SetTextColor(m_hwndListView, crColor);
	// force update
	InvalidateRect(m_hwndListView, NULL, TRUE);
	UpdateWindow(m_hwndListView);
}

void CGroup::SetTextBkColor(COLORREF crColor) {
	ListView_SetTextBkColor(m_hwndListView, crColor);
	// force update
	InvalidateRect(m_hwndListView, NULL, TRUE);
	UpdateWindow(m_hwndListView);
}

void CGroup::SetBackgroundColor(COLORREF crColor) {
	ListView_SetBkColor(m_hwndListView, crColor);
	// force update
	InvalidateRect(m_hwndListView, NULL, TRUE);
	UpdateWindow(m_hwndListView);
}

void CGroup::SetBackground(const char *szBackground, int nOffsetPercentX, int nOffsetPercentY, bool bTiled) {
	LVBKIMAGE lvbki;
	lvbki.ulFlags = LVBKIF_SOURCE_URL | (bTiled ? LVBKIF_STYLE_TILE : LVBKIF_STYLE_NORMAL);
	lvbki.hbm = NULL;
	lvbki.pszImage = const_cast<LPSTR>(szBackground);
	lvbki.xOffsetPercent = nOffsetPercentX;
	lvbki.yOffsetPercent = nOffsetPercentY;
	ListView_SetBkImage(m_hwndListView, &lvbki);
	// force update
	InvalidateRect(m_hwndListView, NULL, TRUE);
	UpdateWindow(m_hwndListView);
}


/**************************************************************************************************
	This function saves the icons current position.
**************************************************************************************************/
void CGroup::SaveState() {
	HKEY hKey;
	DWORD dwDisposition;
	BYTE *pBuf;
	POINT pt;
	LPITEMIDLIST pidl;
	LVITEM lvitem;
	int nCount, nSize, npidlSize;
	char szKeyName[MAX_PATH], szNum[10], szFileName[MAX_PATH];

	if (m_bUseIconPositionFile)
		return SaveToFile();

	StringCchPrintf(szKeyName, MAX_PATH, "%s%s\\%s", REGKEY_MODULE, m_szName, m_szFolderLocation);

	Reset("", true);

	if (ERROR_SUCCESS != RegCreateKeyEx( HKEY_CURRENT_USER, szKeyName, 0, NULL, REG_OPTION_NON_VOLATILE, KEY_ALL_ACCESS, NULL, &hKey, &dwDisposition))
		return;

	// save a string (x,y,pidl) for each item
	nCount = ListView_GetItemCount(m_hwndListView);
	for (int i = 0; i < nCount; i++) {
		lvitem.mask = LVIF_PARAM;
		lvitem.iItem = i;
		lvitem.iSubItem = 0;
		ListView_GetItem(m_hwndListView, &lvitem);
		ListView_GetItemText(m_hwndListView, i, 0, szFileName, MAX_PATH);
		if (!this->IconShouldBeHidden(szFileName)) {
			pidl = (LPITEMIDLIST)(lvitem.lParam);
			ListView_GetItemPosition(m_hwndListView, i, &pt);
			npidlSize = PIDL::GetSize(pidl);
			nSize = 2 * sizeof(ULONG) + npidlSize;
			pBuf = reinterpret_cast<BYTE*>(HeapAlloc(GetProcessHeap(),0, nSize));
			((ULONG*)pBuf)[0] = pt.x;
			((ULONG*)pBuf)[1] = pt.y;
			CopyMemory(pBuf + 2 * sizeof(ULONG), pidl, npidlSize);
			StringCchPrintf(szNum, 10, "%d", i);
			RegSetValueEx(hKey, szNum, 0, REG_BINARY, pBuf, nSize);
			HeapFree(GetProcessHeap(), 0, pBuf);
		}
	}

	RegCloseKey(hKey);
}
/**************************************************************************************************
	This function saves the icons positions into a file. It's called by SaveState()
**************************************************************************************************/
void CGroup::SaveToFile() {
	POINT pt;
	char szFileData[MAX_LINE_LENGTH] = {0}, szFileName[MAX_PATH], szPoint[20], szFile[MAX_PATH];
	int nCount = ListView_GetItemCount(m_hwndListView), iPos = 0;
	UINT j;
	size_t FileNameSize, PointSize;

	if (!(nCount > 0))
		return;

	for (int i = 0; i < nCount; i++) {
		ListView_GetItemText(m_hwndListView, i, 0, szFileName, MAX_PATH);
		if (!this->IconShouldBeHidden(szFileName)) {
			ListView_GetItemPosition(m_hwndListView, i, &pt);
			StringCchLength(szFileName, MAX_PATH, &FileNameSize);
			for(j = 0; j < FileNameSize; j++)
				szFileData[iPos++] = szFileName[j];
			StringCchPrintf(szPoint, 20, "=?%d?%d", pt.x, pt.y);
			StringCchLength(szPoint, MAX_PATH, &PointSize);
			for(j = 0; j <= PointSize; j++)
				szFileData[iPos++] = szPoint[j];
		}
	}
	strcpy_s(szFile, m_szFolderLocation);
	if (szFile[0] == m_nCustomDrive)
		szFile[0] = 63;

	WritePrivateProfileSection(szFile, szFileData, m_szIconPositionFile);
}

/**************************************************************************************************
	This function restores saved icon positions
**************************************************************************************************/
void CGroup::RestoreState() {
	HKEY hKey;
	BYTE *pBuf;
	POINT pt;
	LPITEMIDLIST pidl;
	LVITEM lvitem;
	int nCount;
	char szKeyName[MAX_PATH], szNum[10];
	DWORD dwSize;

	if (m_bUseIconPositionFile)
		return RestoreFromFile();

	// open the registry key
	StringCchPrintf(szKeyName, MAX_PATH, "%s%s\\%s", REGKEY_MODULE, m_szName, m_szFolderLocation);

	if (ERROR_SUCCESS != RegOpenKeyEx( HKEY_CURRENT_USER, szKeyName, 0, KEY_ALL_ACCESS, &hKey))
		return;

	// don't snap icons to grid
	ListView_SetExtendedListViewStyle(m_hwndListView, 0);

	// for each record in the registry key...
	for (int k = 0; ; k++) {
		StringCchPrintf(szNum, 10, "%d", k);
		if (RegQueryValueEx(hKey, szNum, NULL, NULL, NULL, &dwSize) != ERROR_SUCCESS)
			break;

		pBuf = reinterpret_cast<BYTE*>(HeapAlloc(GetProcessHeap(),0, dwSize));

		if (RegQueryValueEx(hKey, szNum, NULL, NULL, pBuf, &dwSize) != ERROR_SUCCESS)
			break;

		pt.x = ((ULONG*)pBuf)[0];
		pt.y = ((ULONG*)pBuf)[1];

		pidl = (LPITEMIDLIST)(pBuf + 2 * sizeof(ULONG));

		// try to find this pidl in the current FolderView

		nCount = ListView_GetItemCount(m_hwndListView);
		for (int i = 0; i < nCount; i++) {
			lvitem.mask = LVIF_PARAM;
			lvitem.iItem = i;
			lvitem.iSubItem = 0;
			ListView_GetItem(m_hwndListView, &lvitem);

			if (lvitem.lParam && m_pFolder->CompareIDs(0, pidl, (LPITEMIDLIST)(lvitem.lParam)) == 0) {
				ListView_SetItemPosition(m_hwndListView, i, pt.x, pt.y);
				break;
			}
		}
		HeapFree(GetProcessHeap(), 0, pBuf);
	}

	RegCloseKey(hKey);
	DWORD dListStyleFlags = LVS_EX_TRANSPARENTBKGND | LVS_EX_LABELTIP;
	if (!m_bNoInfoTips)
		dListStyleFlags |= LVS_EX_INFOTIP;
	if (m_bHotTracking)
		dListStyleFlags |= LVS_EX_TRACKSELECT;
	if (m_bSnapToGrid)
		dListStyleFlags |= LVS_EX_SNAPTOGRID;
	ListView_SetExtendedListViewStyle(m_hwndListView, dListStyleFlags);

	if (m_bSnapToGrid)
		ListView_Arrange(m_hwndListView, LVA_SNAPTOGRID);
}

/**************************************************************************************************
	This function restores saved icons from a file. It's called by RestoreState()
**************************************************************************************************/
void CGroup::RestoreFromFile() {
	char szStoredPos[MAX_LINE_LENGTH], * pString, * pSub, *next;
	char szFileName[MAX_PATH], szTempName[MAX_PATH], szFile[MAX_PATH];
	POINT pt;
	size_t pos = 0;
	int nCount = ListView_GetItemCount(m_hwndListView);

	strcpy_s(szFile, m_szFolderLocation);
	if (szFile[0] == m_nCustomDrive)
		szFile[0] = 63;

	GetPrivateProfileSection(szFile, szStoredPos, MAX_LINE_LENGTH, m_szIconPositionFile);

	ListView_SetExtendedListViewStyle(m_hwndListView, 0);

	pString = szStoredPos + pos;
	while(pString[0] != 0) {
		// This will get the file name
		pSub = strtok_s(pString, "?", &next);
		strcpy_s(szFileName, pSub);
		szFileName[strlen(szFileName)-1] = 0;
		
		// This will get the x coordinate
		pos += (strlen(pString)+1);
		pString = szStoredPos + pos;
		pSub = strtok_s(pString, "?", &next);
		pt.x = atol(pSub);

		// This will get the y coordinate
		pos += (strlen(pString)+1);
		pString = szStoredPos + pos;
		pSub = strtok_s(pString, "?", &next);
		pt.y = atol(pSub);

		for (int i = 0; i < nCount; i++) {
			ListView_GetItemText(m_hwndListView, i, 0, szTempName, MAX_PATH);
			if (_stricmp(szTempName, szFileName) == 0) {
				ListView_SetItemPosition(m_hwndListView, i, pt.x, pt.y);
				break;
			}
		}
		pos += (strlen(pString)+1);
		pString = szStoredPos + pos;
	}
	DWORD dListStyleFlags = LVS_EX_TRANSPARENTBKGND | LVS_EX_LABELTIP;
	if (!m_bNoInfoTips)
		dListStyleFlags |= LVS_EX_INFOTIP;
	if (m_bHotTracking)
		dListStyleFlags |= LVS_EX_TRACKSELECT;
	if (m_bSnapToGrid)
		dListStyleFlags |= LVS_EX_SNAPTOGRID;
	ListView_SetExtendedListViewStyle(m_hwndListView, dListStyleFlags);
	if (m_bSnapToGrid)
		ListView_Arrange(m_hwndListView, LVA_SNAPTOGRID);
}


//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// Drag&Drop support


IDropTarget* CGroup::GetLastItemTarget(bool *pbExecute) {
	IDropTarget* rv = NULL;
	*pbExecute = false;
	if (m_nLastItem != -1) {
		LVITEM lvitem;
		lvitem.mask = LVIF_PARAM;
		lvitem.iItem = m_nLastItem;
		lvitem.iSubItem = 0;
		ListView_GetItem(m_hwndListView, &lvitem);

		if (lvitem.lParam) {
			LPITEMIDLIST pidl = (LPITEMIDLIST)(lvitem.lParam);
			SFGAOF flags = SFGAO_DROPTARGET;
			m_pFolder->GetAttributesOf(1, (LPCITEMIDLIST*)&pidl, &flags);
			bool bIsDropTarget = (flags & SFGAO_DROPTARGET) != 0;
			flags = SFGAO_FILESYSTEM;
			m_pFolder->GetAttributesOf(1, (LPCITEMIDLIST*)&pidl, &flags);
			bool bIsFileSystemObject = 0 != (flags & SFGAO_FILESYSTEM);
			flags = SFGAO_FOLDER;
			m_pFolder->GetAttributesOf(1, (LPCITEMIDLIST*)&pidl, &flags);
			bool bIsFolder = 0 != (flags & SFGAO_FOLDER);
			// file system objects except folders should be managed another way...
			if (bIsFileSystemObject && !bIsFolder)
				*pbExecute = true;
			else if (bIsDropTarget)
				m_pFolder->GetUIObjectOf(m_hWnd, 1, (LPCITEMIDLIST*)&pidl, IID_IDropTarget, NULL, (void**)&rv);
		}
	}
	return rv;
}


void CGroup::Drop(DWORD grfKeyState, POINTL *ptl, DWORD *pdwEffect, CIDA *pida, POINT *pOffsets) {
	LPITEMIDLIST pidl;	// single file object
	UINT i = 0;

	// file dropped on the icon
	if (m_nLastItem != -1) {
		LVITEM lvitem;
		lvitem.mask = LVIF_PARAM;
		lvitem.iItem = m_nLastItem;
		lvitem.iSubItem = 0;
		ListView_GetItem(m_hwndListView, &lvitem);

		bool bMoveOnly = false;
		if (lvitem.lParam) {
			pidl = (LPITEMIDLIST)(lvitem.lParam);

			for (UINT j = 0; j < pida->cidl; j++) {
				if (m_pFolder->CompareIDs(0, pidl, HIDA_GetPIDLItem(pida, j)) == 0) {
					bMoveOnly = true;
					break;
				}
			}

			if (!bMoveOnly) {
				char szFileName[MAX_PATH] = {0};
				LPITEMIDLIST pidlFull = PIDL::Append(m_pidlFolder, pidl);
				if (SHGetPathFromIDList(pidlFull, szFileName)) {
					CoTaskMemFree(pidlFull);
					while (i < pida->cidl) {
						char szParameter[MAX_PATH] = {0};
						pidlFull = PIDL::Append(
							HIDA_GetPIDLParent(pida),
							HIDA_GetPIDLItem(pida, i));
						if (SHGetPathFromIDList(pidlFull, szParameter)) {
							PathQuoteSpaces(szParameter);
							ShellExecute(m_hWnd, NULL, szFileName, szParameter, NULL, SW_SHOWNORMAL);
						}
						CoTaskMemFree(pidlFull);
						++i;
					}
				}
			}
		}

		ListView_SetItemState(m_hwndView, -1, 0, LVIS_DROPHILITED);
		if (!bMoveOnly) {
			*pdwEffect = DROPEFFECT_COPY;
			return;
		}
	}

	while (i < pida->cidl) {
		// we need just to reorder icons...
		int nCount = ListView_GetItemCount(m_hwndListView);
		for (int j = 0; j < nCount; j++) {
			LVITEM lvitem;
			lvitem.mask = LVIF_PARAM;
			lvitem.iItem = j;
			lvitem.iSubItem = 0;
			ListView_GetItem(m_hwndListView, &lvitem);
			if (lvitem.lParam && m_pFolder->CompareIDs(0, HIDA_GetPIDLItem(pida, i), (LPITEMIDLIST)lvitem.lParam) == 0) {
				RECT rect;
				GetWindowRect(m_hwndListView, &rect);
				ListView_SetItemPosition(
					m_hwndListView, j,
					ptl->x + pOffsets[1 + i].x - rect.left,
					ptl->y + pOffsets[1 + i].y - rect.top);
				//m_pView2->SelectAndPositionItem(HIDA_GetPIDLItem(pida, i), SVSI_SELECT | SVSI_TRANSLATEPT, &pt);
				break;
			}
		}
		++i;
	}
	ClearDropHover();
}

void CGroup::SetDropHover(POINTL *pt, DWORD *pdwEffect) {
	// Damn, FindItem doesn't work =(

	RECT rect;

	GetWindowRect(m_hwndListView, &rect);
	pt->x -= rect.left;
	pt->y -= rect.top;

	if (m_nLastItem != -1) {
		ListView_GetItemRect(m_hwndListView, m_nLastItem, &rect, LVIR_BOUNDS);
		if (!(rect.left <= pt->x && pt->x < rect.right && rect.top <= pt->y && pt->y <= rect.bottom)) {
			ListView_SetItemState(m_hwndListView, m_nLastItem, 0, LVIS_DROPHILITED);
			m_nLastItem = -1;
		}
	}

	if (m_nLastItem == -1) {
		int nItems = ListView_GetItemCount(m_hwndListView);
		for (int i = 0; i < nItems; i++) {
			ListView_GetItemRect(m_hwndListView, i, &rect, LVIR_BOUNDS);
			if (rect.left <= pt->x && pt->x < rect.right && rect.top <= pt->y && pt->y <= rect.bottom) {
				m_nLastItem = i;
				ListView_SetItemState(m_hwndListView, m_nLastItem, LVIS_DROPHILITED, LVIS_DROPHILITED);
				break;
			}
		}
	}
	
	if (m_nLastItem != -1) {
		LVITEM lvitem;
		lvitem.mask = LVIF_PARAM;
		lvitem.iItem = m_nLastItem;
		lvitem.iSubItem = 0;
		ListView_GetItem(m_hwndListView, &lvitem);

		if (lvitem.lParam) {
			SFGAOF flags = SFGAO_DROPTARGET;
			m_pFolder->GetAttributesOf(1, (LPCITEMIDLIST*)(&(lvitem.lParam)), &flags);
			if ((flags & SFGAO_DROPTARGET) != SFGAO_DROPTARGET)
				*pdwEffect = DROPEFFECT_NONE;
		}
	}
}

void CGroup::ClearDropHover() {
	if (m_nLastItem != -1)
		ListView_SetItemState(m_hwndListView, m_nLastItem, 0, LVIS_DROPHILITED);
}

/**************************************************************************************************
	This function sets up the list view
**************************************************************************************************/
bool CGroup::InitFolderView() {
	m_pFolder = PIDL::GetFolder(m_pidlFolder);
	if (m_pFolder == NULL)
		return false;
	m_pFolder->CreateViewObject(m_hWnd, IID_IShellView, (void**)&m_pView);
	if (!m_pView)
		return false;

	FOLDERSETTINGS fs;
	fs.ViewMode = m_uViewMode; // FVM_ICON, FVM_SMALLICON, FVM_LIST, FVM_DETAILS, FVM_THUMBNAIL, FVM_TILE, FVM_THUMBSTRIP
	fs.fFlags = FWF_NOCLIENTEDGE | FWF_SNAPTOGRID | FWF_SHOWSELALWAYS /*| FWF_TRANSPARENT | FWF_DESKTOP | FWF_BESTFITWINDOW*/;
	if (!m_szBackground || !(m_szBackground[0])) {
		fs.fFlags |= FWF_TRANSPARENT /*| FWF_NOWEBVIEW*/ | FWF_DESKTOP;
	}
	else
		fs.fFlags |= FWF_NOWEBVIEW;

	if (m_bNoScroll)
		fs.fFlags |= FWF_NOSCROLL;

	if (m_bAutoArrange)
		fs.fFlags |= FWF_AUTOARRANGE;
	
	RECT rect = {
		m_bDestroyParent ? 0 : m_nX,
		m_bDestroyParent ? 0 : m_nY,
		m_bDestroyParent ? m_nWidth : m_nX + m_nWidth,
		m_bDestroyParent ? m_nHeight : m_nY + m_nHeight
	};

	m_pView->CreateViewWindow(NULL, &fs, (IShellBrowser*)this, &rect, &m_hwndView);
	m_pView->UIActivate(SVUIA_ACTIVATE_NOFOCUS);
	if (!m_hwndView)
		return false;

	SetWindowPos(m_hwndView, HWND_BOTTOM, 0,0,0,0, SWP_NOACTIVATE | SWP_NOMOVE | SWP_NOSIZE);

	m_pView->QueryInterface(IID_IShellView2, (void**)&m_pView2);
	if (m_pView2 == NULL)
		return false;

	m_hwndListView = ::GetWindow(m_hwndView, GW_CHILD);

	m_longListViewStyle = GetWindowLong(m_hwndListView, GWL_STYLE);

	if (m_bAlignTop)
		m_longListViewStyle ^= LVS_ALIGNLEFT;

	if (m_bAlwaysShowSelection)
		m_longListViewStyle |= LVS_SHOWSELALWAYS;

	SetWindowLongPtr(m_hwndListView, GWL_STYLE, m_longListViewStyle);

	SetWindowLongPtr(m_hwndView, GWLP_USERDATA, (LONG)this);
	wpOrigGroup_Proc = (WNDPROC) SetWindowLongPtr(m_hwndView, GWLP_WNDPROC, (LONG) Group_Proc);

	m_pDropTarget = new CDropTarget();
	m_pDropTarget->Register(this);

	if (m_hFont)
		SendMessage(m_hwndListView, WM_SETFONT, (WPARAM)m_hFont, MAKELPARAM(TRUE, 0));

	ListView_SetTextBkColor(m_hwndListView,m_crTextBackgroundColor);
	ListView_SetTextColor(m_hwndListView, m_crTextColor);
	ListView_SetIconSpacing(m_hwndListView, m_nSpacingX, m_nSpacingY);
	ListView_SetBkColor(m_hwndListView, m_crBackgroundColor);

	if (m_uViewMode == FVM_SMALLICON)
		ListView_SetView(m_hwndListView, LV_VIEW_SMALLICON);

	if (m_szBackground && m_szBackground[0])
		SetBackground(m_szBackground, m_nOffsetPercentX, m_nOffsetPercentY, m_bTiled);

	// Hijack the window procedure for the listview.
	SetWindowLongPtr(m_hwndListView, GWLP_USERDATA, (LONG)this);
	wpOrigListView_Proc = (WNDPROC) SetWindowLongPtr(m_hwndListView, GWLP_WNDPROC, (LONG) ListView_Proc);

	ScanIcons();
	RestoreState();

	// Export Evars
	char szEvar[MAX_RCCOMMAND], szValue[MAX_RCCOMMAND];
	sprintf_s(szEvar, MAX_RCCOMMAND, "%s%s", m_szName, "CurrentX");
	_itoa_s(m_nX, szValue, MAX_RCCOMMAND, 10);
	LSSetVariable(szEvar, szValue);
	sprintf_s(szEvar, MAX_RCCOMMAND, "%s%s", m_szName, "CurrentY");
	_itoa_s(m_nY, szValue, MAX_RCCOMMAND, 10);
	LSSetVariable(szEvar, szValue);
	sprintf_s(szEvar, MAX_RCCOMMAND, "%s%s", m_szName, "CurrentWidth");
	_itoa_s(m_nWidth, szValue, MAX_RCCOMMAND, 10);
	LSSetVariable(szEvar, szValue);
	sprintf_s(szEvar, MAX_RCCOMMAND, "%s%s", m_szName, "CurrentHeight");
	_itoa_s(m_nHeight, szValue, MAX_RCCOMMAND, 10);
	LSSetVariable(szEvar, szValue);	
	sprintf_s(szEvar, MAX_RCCOMMAND, "%s%s", m_szName, "ItemCount");
	_itoa_s(ListView_GetItemCount(m_hwndListView), szValue, 10, 10);
	LSSetVariable(szEvar, szValue);

	ShowWindow(m_hwndView, m_bHidden ? SW_HIDE : SW_SHOW);

	return true;
}

/**************************************************************************************************
	This function forwards mouse messages to m_hWnd when they aren't handled by clickonic.
**************************************************************************************************/
void CGroup::relayMouseMessage(UINT message, WPARAM wParam, LPARAM lParam) {
	POINT pt = { GET_X_LPARAM(lParam), GET_Y_LPARAM(lParam) };
	MapWindowPoints( ((message == WM_MOUSEWHEEL) || (message == WM_MOUSEHWHEEL)) ? HWND_DESKTOP : m_hwndListView, m_hWnd, &pt, 1);
	PostMessage(m_hWnd, message, wParam, MAKELPARAM((short) pt.x, (short) pt.y));
}

/**************************************************************************************************
	This function finds or creates a parent window and returns its hWnd.
**************************************************************************************************/
HWND CGroup::GetParentWindow() {
	m_bDestroyParent = false; // We need to remember whether or not we should destroy the parent hwnd

	// This is the prefered scenario, mousemessages will be forwarded to m_hWnd.
	HWND hWnd = FindWindow("DesktopBackgroundClass", NULL);
	if (hWnd)
		return hWnd;

	// Well, we didn't find any desktop modules, so let's create our own window.
	hWnd = CreateWindowEx(WS_EX_TOOLWINDOW, g_szWClass, m_szName, WS_POPUP|WS_CLIPSIBLINGS|WS_CLIPCHILDREN,
		m_nX, m_nY, m_nWidth, m_nHeight, NULL, NULL, g_hInstance, NULL);

	if (!hWnd)
		return GetDesktopWindow();// Geez, nothing is working... This will display the icons at any rate.

	m_bDestroyParent = true;
	SetWindowLongPtr(hWnd, GWLP_USERDATA, (LONG)this);
	SetWindowPos(hWnd, HWND_BOTTOM, 0,0,0,0, SWP_NOSIZE|SWP_NOMOVE|SWP_NOACTIVATE|SWP_NOSENDCHANGING);
	ShowWindow(hWnd, SW_SHOWNOACTIVATE);

	return hWnd;
}

/**************************************************************************************************
	This function should change the my documens path from it's real path to it's CLSID.
**************************************************************************************************/
/*void CGroup::FixMyDocs() {
	STRRET strDispName;
	char szFileName[MAX_PATH];
	size_t FileNameSize;
	LVITEM lvitem;
	int nCount = ListView_GetItemCount(m_hwndListView);
	char szMyDocs[] = "::{450D8FBA-AD25-11D0-98A8-0800361B1103}";
	WCHAR wMyDocs[MAX_PATH];
	ITEMIDLIST* id;
	ULONG ulEaten;

	MultiByteToWideChar(CP_ACP, 0, szMyDocs, -1, wMyDocs, MAX_PATH);
	
	//SHILCreateFromPath(wMyDocs, &id, NULL);
	//id = SHSimpleIDListFromPath(wMyDocs);

	m_pFolder->ParseDisplayName(m_hwndListView, NULL, wMyDocs, &ulEaten, &id, NULL);
	//m_pFolder->BindToObject(

	sprintf_s(szFileName, MAX_PATH, "%s", id);

	for (int i = 0; i < sizeof(szFileName); i++)
		if (szFileName[i] == 0)
			szFileName[i] = 63;

	MessageBox(NULL, szFileName, "Debug - Filename", MB_OK);
	
	/*for (int i = 0; i < nCount; i++) {
		ZeroMemory(&lvitem, sizeof(LVITEM));
		lvitem.mask = LVIF_PARAM;
		lvitem.iItem = i;
		lvitem.iSubItem = 0;
		ListView_GetItem(m_hwndListView, &lvitem);
		m_pFolder->GetDisplayNameOf((LPCITEMIDLIST)(lvitem.lParam), SHGDN_FORPARSING, &strDispName);
		StrRetToBuf(&strDispName, (LPCITEMIDLIST)(lvitem.lParam), szFileName, MAX_PATH);
		StringCchLength(szFileName, MAX_PATH, &FileNameSize);
		
		if (!_stricmp(szFileName, m_szMyDocumentsPath)) {
			m_pFolder->SetNameOf(
				m_hwndListView,
				(LPCITEMIDLIST)(lvitem.lParam),
				wMyDocs,
				SHGDN_FORPARSING,
				&id);
		}
	}
}*/

/**************************************************************************************************
	This function fires when a "Windows" User Policy changes (such as desktop icon settings)
**************************************************************************************************/
/*void CGroup::HandleSettingChange() {
	// We should make this function restore hidden icons as well...
	if (m_bDontUseRegistryIcons)
		return;
	ReadIconSettings();
	ScanIcons();
}*/

/**************************************************************************************************
	This function takes a listview item id and returns it's full parsename.
**************************************************************************************************/
void CGroup::GetNameFromId(int iItem, char* szFileName) {
	STRRET strDispName;
	LVITEM lvitem;

	ZeroMemory(&lvitem, sizeof(LVITEM));
	lvitem.mask = LVIF_PARAM;
	lvitem.iItem = iItem;
	lvitem.iSubItem = 0;
	ListView_GetItem(m_hwndListView, &lvitem);
	m_pFolder->GetDisplayNameOf((LPCITEMIDLIST)(lvitem.lParam), SHGDN_FORPARSING, &strDispName);
	StrRetToBuf(&strDispName, (LPCITEMIDLIST)(lvitem.lParam), szFileName, MAX_PATH);
}

/**************************************************************************************************
	This function takes user input and turns it into a modkey word.
**************************************************************************************************/
WORD CGroup::GetModkeyWord(char* szOption, char* szDefault) {
	WORD wReturn = 0;
	char szBuf[MAX_LINE_LENGTH];
	LiteStep::GetPrefixedRCLine(szBuf, m_szName, szOption, szDefault);
	if (strstr(szBuf, ".shift") != NULL)
		wReturn |= MODKEY_SHIFT;
	if (strstr(szBuf, ".alt") != NULL)
		wReturn |= MODKEY_ALT;
	if (strstr(szBuf, ".ctrl") != NULL)
		wReturn |= MODKEY_CTRL;
	return wReturn;
}

/**************************************************************************************************
	This function handles events.
**************************************************************************************************/
void CGroup::HandleEvent(UINT uEvent) {
	if (_stricmp(m_szEventCommand[uEvent], EVENT_ACTION_NONE) == 0)
		return;
	else if (_stricmp(m_szEventCommand[uEvent], EVENT_ACTION_RENAME) == 0) {
		ListView_EditLabel(m_hwndListView, ListView_GetNextItem(m_hwndListView, -1, LVNI_FOCUSED));
	}
	else if (_stricmp(m_szEventCommand[uEvent], EVENT_ACTION_EXECUTE) == 0) {
		if (m_bInlineBrowsing && ListView_GetSelectedCount(m_hwndListView) == 1) {
			int iItem = ListView_GetNextItem(m_hwndListView, -1, LVNI_SELECTED);
			char szFileName[MAX_PATH_LENGTH];
			GetNameFromId(iItem, szFileName);
			if (utils::Is_Directory(szFileName))
				 return ChangeDir(".selected");
		}
		CallWindowProc(wpOrigListView_Proc, m_hwndListView, WM_KEYDOWN, VK_RETURN, NULL);
	}
	else if (_stricmp(m_szEventCommand[uEvent], EVENT_ACTION_DELETE) == 0) {
		DeleteFiles();
	}
	else if (_stricmp(m_szEventCommand[uEvent], EVENT_ACTION_PASTE) == 0) {
		DoPaste();
	}
	else if (_stricmp(m_szEventCommand[uEvent], EVENT_ACTION_UNDO) == 0) {
		// We need to add support for this
	}
	else if (_stricmp(m_szEventCommand[uEvent], EVENT_ACTION_COPY) == 0) {
		CopyFiles(false);
	}
	else if (_stricmp(m_szEventCommand[uEvent], EVENT_ACTION_CUT) == 0) {
		CopyFiles(true);
	}
	else if (_stricmp(m_szEventCommand[uEvent], EVENT_ACTION_SELECTALL) == 0) {
		ListView_SetItemState(m_hwndListView, -1, LVIS_SELECTED, LVIS_SELECTED);
	}
	else if (_stricmp(m_szEventCommand[uEvent], EVENT_ACTION_REFRESH) == 0) {
		SaveState();
		m_pView2->Refresh();
		RestoreState();
	}
	else if (_stricmp(m_szEventCommand[uEvent], EVENT_ACTION_UP) == 0) {
		ChangeDir(".up");
	}
	else if (_stricmp(m_szEventCommand[uEvent], EVENT_ACTION_CHANGEDIR) == 0) {
		ChangeDir(".selected");
	}
	else if (_stricmp(m_szEventCommand[uEvent], EVENT_ACTION_DOWNKEY) == 0) {
		CallWindowProc(wpOrigListView_Proc, m_hwndListView, WM_KEYDOWN, VK_DOWN, NULL);
	}
	else if (_stricmp(m_szEventCommand[uEvent], EVENT_ACTION_UPKEY) == 0) {
		CallWindowProc(wpOrigListView_Proc, m_hwndListView, WM_KEYDOWN, VK_UP, NULL);
	}
	else if (_stricmp(m_szEventCommand[uEvent], EVENT_ACTION_LEFTKEY) == 0) {
		CallWindowProc(wpOrigListView_Proc, m_hwndListView, WM_KEYDOWN, VK_LEFT, NULL);
	}
	else if (_stricmp(m_szEventCommand[uEvent], EVENT_ACTION_RIGHTKEY) == 0) {
		CallWindowProc(wpOrigListView_Proc, m_hwndListView, WM_KEYDOWN, VK_RIGHT, NULL);
	}
	else {
		LSExecute(m_hwndListView, m_szEventCommand[uEvent], SW_SHOWNORMAL);
	}
}

/**************************************************************************************************
	This function handles mouse events.
**************************************************************************************************/
void CGroup::HandleMouseEvent(UINT uEvent, UINT msg, WPARAM wParam, LPARAM lParam) {
	if (m_bExplorerDesktop) {
		CallWindowProc(wpOrigListView_Proc, m_hwndListView, msg, wParam, lParam);
		return;
	}
	if (msg == WM_LBUTTONDOWN || msg == WM_RBUTTONDOWN || msg == WM_LBUTTONDBLCLK) {
		LVHITTESTINFO lvhi;
		ZeroMemory(&lvhi, sizeof(LVHITTESTINFO));
		lvhi.flags = LVHT_ONITEM;
		lvhi.pt.x = GET_X_LPARAM(lParam);
		lvhi.pt.y = GET_Y_LPARAM(lParam);

		if ( ListView_HitTest(m_hwndListView, &lvhi) != -1 ) {
			// We could add a whole set of events for this scenario as well, but I don't really see a need.
			CallWindowProc(wpOrigListView_Proc, m_hwndListView, msg, wParam, lParam);
			return; // An icon was clicked.
		}
	}

	if (m_bStealFocus)
		SetFocus(m_hwndListView);

	if (_stricmp(m_szEventCommand[uEvent], EVENT_ACTION_CHANGEDIR) == 0) {
		// We can override HandleEvent actions like this ;)
		LVHITTESTINFO lvhi;
		ZeroMemory(&lvhi, sizeof(LVHITTESTINFO));
		lvhi.flags = LVHT_ONITEM;
		lvhi.pt.x = GET_X_LPARAM(lParam);
		lvhi.pt.y = GET_Y_LPARAM(lParam);

		if (msg == WM_MOUSEWHEEL || msg == WM_MOUSEHWHEEL)
			MapWindowPoints(HWND_DESKTOP, m_hwndListView, &lvhi.pt, 1);
			
		int nFileid = ListView_HitTest(m_hwndListView, &lvhi);
		if ( nFileid != -1 ) {
			char szFileName[MAX_PATH_LENGTH];
			GetNameFromId(nFileid, szFileName);
			if (utils::Is_Directory(szFileName)) {
				if (!m_bNoVirtualSwitch && (_stricmp(szFileName, m_szDesktopPath) == 0))
					ChangeDir(".desktop");
				else
					ChangeDir(szFileName);
				return;
			}
		}
	}

	// We just need to handle the mouse specific events and then forward the event to HandleEvent
	if (_stricmp(m_szEventCommand[uEvent], EVENT_ACTION_RECTANGLE) == 0) {
		if (CheckModkeys(m_wSelectionRectangleMod))
			CallWindowProc(wpOrigListView_Proc, m_hwndListView, WM_LBUTTONDOWN, wParam, lParam);
		else
			relayMouseMessage(msg, wParam, lParam);
	}
	else if (_stricmp(m_szEventCommand[uEvent], EVENT_ACTION_SYSTEMMENU) == 0) {
		if (CheckModkeys(m_wSystemMenuMod))
			CallWindowProc(wpOrigListView_Proc, m_hwndListView, WM_RBUTTONDOWN, wParam, lParam);
		else
			relayMouseMessage(msg, wParam, lParam);
	}
	else if (_stricmp(m_szEventCommand[uEvent], EVENT_ACTION_FORWARD) == 0) {
		relayMouseMessage(msg, wParam, lParam);
	}
	else {
		HandleEvent(uEvent);
	}
}

/**************************************************************************************************
	This function loads commands for all events
**************************************************************************************************/
void CGroup::LoadEvents() {
	// Backwards compability
	bool bUseDelete	= LiteStep::GetPrefixedRCBool(m_szName, "UseDelete", TRUE);
	bool bUseRename	= LiteStep::GetPrefixedRCBool(m_szName, "UseRename", TRUE);
	bool bDontUseCtrlA = LiteStep::GetPrefixedRCBool(m_szName, "DontUseCtrlA", FALSE);
	bool bDontUseCtrlC = LiteStep::GetPrefixedRCBool(m_szName, "DontUseCtrlC", FALSE);
	bool bDontUseCtrlV = LiteStep::GetPrefixedRCBool(m_szName, "DontUseCtrlV", FALSE);
	bool bDontUseCtrlX = LiteStep::GetPrefixedRCBool(m_szName, "DontUseCtrlX", FALSE);
	bool bUseSelectionRectangle = LiteStep::GetPrefixedRCBool(m_szName, "UseSelectionRectangle", FALSE);
	bool bUseSystemMenu = LiteStep::GetPrefixedRCBool(m_szName, "UseSystemMenu", FALSE);
	
	// Event Settings
	LiteStep::GetPrefixedRCLine(m_szEventCommand[EVENT_ONF1DOWN], m_szName, "OnF1Down", EVENT_ACTION_NONE);
	LiteStep::GetPrefixedRCLine(m_szEventCommand[EVENT_ONF2DOWN], m_szName, "OnF2Down", bUseRename ? EVENT_ACTION_RENAME : EVENT_ACTION_NONE);
	LiteStep::GetPrefixedRCLine(m_szEventCommand[EVENT_ONF3DOWN], m_szName, "OnF3Down", EVENT_ACTION_NONE);
	LiteStep::GetPrefixedRCLine(m_szEventCommand[EVENT_ONF4DOWN], m_szName, "OnF4Down", EVENT_ACTION_NONE);
	LiteStep::GetPrefixedRCLine(m_szEventCommand[EVENT_ONF5DOWN], m_szName, "OnF5Down", EVENT_ACTION_REFRESH);
	LiteStep::GetPrefixedRCLine(m_szEventCommand[EVENT_ONF6DOWN], m_szName, "OnF6Down", EVENT_ACTION_NONE);
	LiteStep::GetPrefixedRCLine(m_szEventCommand[EVENT_ONF7DOWN], m_szName, "OnF7Down", EVENT_ACTION_NONE);
	LiteStep::GetPrefixedRCLine(m_szEventCommand[EVENT_ONF8DOWN], m_szName, "OnF8Down", EVENT_ACTION_NONE);
	LiteStep::GetPrefixedRCLine(m_szEventCommand[EVENT_ONF9DOWN], m_szName, "OnF9Down", EVENT_ACTION_NONE);
	LiteStep::GetPrefixedRCLine(m_szEventCommand[EVENT_ONF10DOWN], m_szName, "OnF10Down", EVENT_ACTION_NONE);
	LiteStep::GetPrefixedRCLine(m_szEventCommand[EVENT_ONF11DOWN], m_szName, "OnF11Down", EVENT_ACTION_NONE);
	LiteStep::GetPrefixedRCLine(m_szEventCommand[EVENT_ONF12DOWN], m_szName, "OnF12Down", EVENT_ACTION_NONE);
	LiteStep::GetPrefixedRCLine(m_szEventCommand[EVENT_ONF1UP], m_szName, "OnF1Up", EVENT_ACTION_NONE);
	LiteStep::GetPrefixedRCLine(m_szEventCommand[EVENT_ONF2UP], m_szName, "OnF2Up", EVENT_ACTION_NONE);
	LiteStep::GetPrefixedRCLine(m_szEventCommand[EVENT_ONF3UP], m_szName, "OnF3Up", EVENT_ACTION_NONE);
	LiteStep::GetPrefixedRCLine(m_szEventCommand[EVENT_ONF4UP], m_szName, "OnF4Up", EVENT_ACTION_NONE);
	LiteStep::GetPrefixedRCLine(m_szEventCommand[EVENT_ONF5UP], m_szName, "OnF5Up", EVENT_ACTION_NONE);
	LiteStep::GetPrefixedRCLine(m_szEventCommand[EVENT_ONF6UP], m_szName, "OnF6Up", EVENT_ACTION_NONE);
	LiteStep::GetPrefixedRCLine(m_szEventCommand[EVENT_ONF7UP], m_szName, "OnF7Up", EVENT_ACTION_NONE);
	LiteStep::GetPrefixedRCLine(m_szEventCommand[EVENT_ONF8UP], m_szName, "OnF8Up", EVENT_ACTION_NONE);
	LiteStep::GetPrefixedRCLine(m_szEventCommand[EVENT_ONF9UP], m_szName, "OnF9Up", EVENT_ACTION_NONE);
	LiteStep::GetPrefixedRCLine(m_szEventCommand[EVENT_ONF10UP], m_szName, "OnF10Up", EVENT_ACTION_NONE);
	LiteStep::GetPrefixedRCLine(m_szEventCommand[EVENT_ONF11UP], m_szName, "OnF11Up", EVENT_ACTION_NONE);
	LiteStep::GetPrefixedRCLine(m_szEventCommand[EVENT_ONF12UP], m_szName, "OnF12Up", EVENT_ACTION_NONE);
	LiteStep::GetPrefixedRCLine(m_szEventCommand[EVENT_ONF11UP], m_szName, "OnF11Up", EVENT_ACTION_NONE);
	LiteStep::GetPrefixedRCLine(m_szEventCommand[EVENT_ONF12UP], m_szName, "OnF12Up", EVENT_ACTION_NONE);
	LiteStep::GetPrefixedRCLine(m_szEventCommand[EVENT_ONDELDOWN], m_szName, "OnDELDown", bUseDelete ? EVENT_ACTION_DELETE : EVENT_ACTION_NONE);
	LiteStep::GetPrefixedRCLine(m_szEventCommand[EVENT_ONDELUP], m_szName, "OnDELUp", EVENT_ACTION_NONE);
	LiteStep::GetPrefixedRCLine(m_szEventCommand[EVENT_ONFOLDERCHANGE], m_szName, "OnFolderChange", EVENT_ACTION_NONE);
	LiteStep::GetPrefixedRCLine(m_szEventCommand[EVENT_ONINSERTITEM], m_szName, "OnRemoveItem", EVENT_ACTION_NONE);
	LiteStep::GetPrefixedRCLine(m_szEventCommand[EVENT_ONREMOVEITEM], m_szName, "OnInsertItem", EVENT_ACTION_NONE);
	LiteStep::GetPrefixedRCLine(m_szEventCommand[EVENT_ONCTRLA], m_szName, "OnCtrlA", bDontUseCtrlA ? EVENT_ACTION_NONE : EVENT_ACTION_SELECTALL);
	LiteStep::GetPrefixedRCLine(m_szEventCommand[EVENT_ONCTRLV], m_szName, "OnCtrlV", bDontUseCtrlV ? EVENT_ACTION_NONE : EVENT_ACTION_PASTE);
	LiteStep::GetPrefixedRCLine(m_szEventCommand[EVENT_ONCTRLX], m_szName, "OnCtrlX", bDontUseCtrlX ? EVENT_ACTION_NONE : EVENT_ACTION_CUT);
	LiteStep::GetPrefixedRCLine(m_szEventCommand[EVENT_ONCTRLC], m_szName, "OnCtrlC", bDontUseCtrlC ? EVENT_ACTION_NONE : EVENT_ACTION_COPY);
	LiteStep::GetPrefixedRCLine(m_szEventCommand[EVENT_ONCTRLZ], m_szName, "OnCtrlZ", EVENT_ACTION_UNDO);
	LiteStep::GetPrefixedRCLine(m_szEventCommand[EVENT_ONLEFTCLICKDOWN], m_szName, "OnLeftClickDown", bUseSelectionRectangle ? EVENT_ACTION_RECTANGLE : EVENT_ACTION_FORWARD);
	LiteStep::GetPrefixedRCLine(m_szEventCommand[EVENT_ONRIGHTCLICKDOWN], m_szName, "OnRightClickDown", bUseSystemMenu ? EVENT_ACTION_SYSTEMMENU : EVENT_ACTION_FORWARD);
	LiteStep::GetPrefixedRCLine(m_szEventCommand[EVENT_ONMIDDLECLICKDOWN], m_szName, "OnMiddleClickDown", EVENT_ACTION_FORWARD);
	LiteStep::GetPrefixedRCLine(m_szEventCommand[EVENT_ONX1CLICKDOWN], m_szName, "OnX1ClickDown", EVENT_ACTION_FORWARD);
	LiteStep::GetPrefixedRCLine(m_szEventCommand[EVENT_ONX2CLICKDOWN], m_szName, "OnX2ClickDown", EVENT_ACTION_FORWARD);
	LiteStep::GetPrefixedRCLine(m_szEventCommand[EVENT_ONLEFTCLICKUP], m_szName, "OnLeftClickUp", EVENT_ACTION_FORWARD);
	LiteStep::GetPrefixedRCLine(m_szEventCommand[EVENT_ONRIGHTCLICKUP], m_szName, "OnRightClickUp", EVENT_ACTION_FORWARD);
	LiteStep::GetPrefixedRCLine(m_szEventCommand[EVENT_ONMIDDLECLICKUP], m_szName, "OnMiddleClickUp", EVENT_ACTION_FORWARD);
	LiteStep::GetPrefixedRCLine(m_szEventCommand[EVENT_ONX1CLICKUP], m_szName, "OnX1ClickUp", EVENT_ACTION_FORWARD);
	LiteStep::GetPrefixedRCLine(m_szEventCommand[EVENT_ONX2CLICKUP], m_szName, "OnX2ClickUp", EVENT_ACTION_FORWARD);
	LiteStep::GetPrefixedRCLine(m_szEventCommand[EVENT_ONLEFTDBLCLICK], m_szName, "OnLeftDoubleClick", EVENT_ACTION_FORWARD);
	LiteStep::GetPrefixedRCLine(m_szEventCommand[EVENT_ONRIGHTDBLCLICK], m_szName, "OnRightDoubleClick", EVENT_ACTION_FORWARD);
	LiteStep::GetPrefixedRCLine(m_szEventCommand[EVENT_ONMIDDLEDBLCLICK], m_szName, "OnMiddleDoubleClick", EVENT_ACTION_FORWARD);
	LiteStep::GetPrefixedRCLine(m_szEventCommand[EVENT_ONX1DBLCLICK], m_szName, "OnX1DoubleClick", EVENT_ACTION_FORWARD);
	LiteStep::GetPrefixedRCLine(m_szEventCommand[EVENT_ONX2DBLCLICK], m_szName, "OnX2DoubleClick", EVENT_ACTION_FORWARD);
	LiteStep::GetPrefixedRCLine(m_szEventCommand[EVENT_ONWHEELUP], m_szName, "OnWheelUp", EVENT_ACTION_FORWARD);
	LiteStep::GetPrefixedRCLine(m_szEventCommand[EVENT_ONWHEELDOWN], m_szName, "OnWheelDown", EVENT_ACTION_FORWARD);
	LiteStep::GetPrefixedRCLine(m_szEventCommand[EVENT_ONWHEELLEFT], m_szName, "OnWheelLeft", EVENT_ACTION_FORWARD);
	LiteStep::GetPrefixedRCLine(m_szEventCommand[EVENT_ONWHEELRIGHT], m_szName, "OnWheelRight", EVENT_ACTION_FORWARD);
	LiteStep::GetPrefixedRCLine(m_szEventCommand[EVENT_ONRIGHTKEYDOWN], m_szName, "OnRightKeyDown", EVENT_ACTION_RIGHTKEY);
	LiteStep::GetPrefixedRCLine(m_szEventCommand[EVENT_ONLEFTKEYDOWN], m_szName, "OnLeftKeyDown", EVENT_ACTION_LEFTKEY);
	LiteStep::GetPrefixedRCLine(m_szEventCommand[EVENT_ONUPKEYDOWN], m_szName, "OnUpKeyDown", EVENT_ACTION_UPKEY);
	LiteStep::GetPrefixedRCLine(m_szEventCommand[EVENT_ONDOWNKEYDOWN], m_szName, "OnDownKeyDown", EVENT_ACTION_DOWNKEY);
	LiteStep::GetPrefixedRCLine(m_szEventCommand[EVENT_ONENTERDOWN], m_szName, "OnEnterDown", EVENT_ACTION_EXECUTE);
	LiteStep::GetPrefixedRCLine(m_szEventCommand[EVENT_ONRIGHTKEYUP], m_szName, "OnRightKeyUp", EVENT_ACTION_NONE);
	LiteStep::GetPrefixedRCLine(m_szEventCommand[EVENT_ONLEFTKEYUP], m_szName, "OnLeftKeyUp", EVENT_ACTION_NONE);
	LiteStep::GetPrefixedRCLine(m_szEventCommand[EVENT_ONUPKEYUP], m_szName, "OnUpKeyUp", EVENT_ACTION_NONE);
	LiteStep::GetPrefixedRCLine(m_szEventCommand[EVENT_ONDOWNKEYUP], m_szName, "OnDownKeyUp", EVENT_ACTION_NONE);
	LiteStep::GetPrefixedRCLine(m_szEventCommand[EVENT_ONENTERUP], m_szName, "OnEnterUp", EVENT_ACTION_NONE);
}